from collections import defaultdict
from itertools import chain
from typing import Iterator, List, Union

from graphql import GraphQLSchema

from .build_schema import build_schema_from_type_definitions
from .resolvers import add_resolve_functions_to_schema


def decompose_maps(resolvers_maps: List[dict]) -> Iterator[tuple]:
    def flatten(rm):
        for key, value in rm.items():
            for resolver_name, resolver in value.items():
                yield (key, resolver_name, resolver)

    return chain.from_iterable(flatten(m) for m in resolvers_maps)


def merge_resolvers(resolver_list: Iterator[tuple]) -> dict:
    output = defaultdict(dict)  # type: dict
    for key, resolver_name, resolver in resolver_list:
        output[key][resolver_name] = resolver
    return output


def join_type_defs(type_defs: List[str]) -> str:
    return "\n\n".join(t.strip() for t in type_defs)


def make_executable_schema(
    type_defs: Union[str, List[str]], resolvers: Union[dict, List[dict]]
) -> GraphQLSchema:
    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    schema = build_schema_from_type_definitions(type_defs)

    if isinstance(resolvers, list):
        add_resolve_functions_to_schema(
            schema, merge_resolvers(decompose_maps(resolvers))
        )
    elif isinstance(resolvers, dict):
        add_resolve_functions_to_schema(schema, resolvers)

    return schema



from .constants import HTTP_STATUS_400_BAD_REQUEST, HTTP_STATUS_405_METHOD_NOT_ALLOWED


class HttpError(Exception):
    status = ""

    def __init__(self, message=None):
        super().__init__()
        self.message = message


class HttpBadRequestError(HttpError):
    status = HTTP_STATUS_400_BAD_REQUEST


class HttpMethodNotAllowedError(HttpError):
    status = HTTP_STATUS_405_METHOD_NOT_ALLOWED


class GraphQLError(Exception):
    pass



from graphql import GraphQLSchema, parse
from graphql.language.ast import (
    Document,
    ObjectTypeDefinition,
    OperationTypeDefinition,
    Name,
    NamedType,
    SchemaDefinition,
)
from graphql.utils.build_ast_schema import build_ast_schema


def build_schema_from_type_definitions(type_defs: str) -> GraphQLSchema:
    document = parse(type_defs)

    if not document_has_schema(document):
        schema_definition = build_default_schema(document)
        document.definitions.append(schema_definition)

    return build_ast_schema(document)


def build_default_schema(document: Document) -> SchemaDefinition:
    defined_types = [
        td.name.value
        for td in document.definitions
        if isinstance(td, ObjectTypeDefinition)
    ]
    operations = []
    if "Query" in defined_types:
        operations.append(
            OperationTypeDefinition("query", type=NamedType(name=Name("Query")))
        )
    if "Mutation" in defined_types:
        operations.append(
            OperationTypeDefinition("mutation", type=NamedType(name=Name("Mutation")))
        )
    if "Subscription" in defined_types:
        operations.append(
            OperationTypeDefinition(
                "subscription", type=NamedType(name=Name("Subscription"))
            )
        )
    return SchemaDefinition(operation_types=operations, directives=[])


def document_has_schema(document: Document) -> bool:
    return any(isinstance(td, SchemaDefinition) for td in document.definitions)



from .build_schema import build_schema_from_type_definitions
from .executable_schema import make_executable_schema
from .resolvers import add_resolve_functions_to_schema, default_resolver, resolve_to
from .wsgi_middleware import GraphQLMiddleware

__all__ = [
    "GraphQLMiddleware",
    "add_resolve_functions_to_schema",
    "build_schema_from_type_definitions",
    "default_resolver",
    "make_executable_schema",
    "resolve_to",
]



#! /usr/bin/env python
from setuptools import setup

CLASSIFIERS = [
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name="ariadne",
    author="Mirumee Software",
    author_email="hello@mirumee.com",
    description="A functional implementation of a Python GraphQL server",
    license="BSD",
    version="0.0.3",
    url="https://github.com/mirumee/ariadne",
    packages=["ariadne"],
    install_requires=["graphql-core>=2.1", "typing>=3.6.0"],
    classifiers=CLASSIFIERS,
    platforms=["any"],
)



from graphql import GraphQLObjectType, GraphQLScalarType, GraphQLSchema
from graphql.execution.base import ResolveInfo


def resolve_parent_field(parent, name: str, **kwargs: dict):
    if isinstance(parent, dict):
        value = parent.get(name)
    else:
        value = getattr(parent, name, None)
    if callable(value):
        return value(**kwargs)
    return value


def default_resolver(parent, info: ResolveInfo, **kwargs):
    return resolve_parent_field(parent, info.field_name, **kwargs)


def resolve_to(name: str):
    def resolver(parent, *_, **kwargs):
        return resolve_parent_field(parent, name, **kwargs)

    return resolver


def add_resolve_functions_to_schema(schema: GraphQLSchema, resolvers: dict):
    for type_name, type_object in schema.get_type_map().items():
        if isinstance(type_object, GraphQLObjectType):
            add_resolve_functions_to_object(type_name, type_object, resolvers)
        if isinstance(type_object, GraphQLScalarType):
            add_resolve_functions_to_scalar(type_name, type_object, resolvers)


def add_resolve_functions_to_object(name: str, obj: GraphQLObjectType, resolvers: dict):
    type_resolvers = resolvers.get(name, {})
    for field_name, field_object in obj.fields.items():
        field_resolver = type_resolvers.get(field_name)
        if field_resolver:
            field_object.resolver = field_resolver
        elif field_object.resolver is None:
            field_object.resolver = default_resolver


def add_resolve_functions_to_scalar(name: str, obj: GraphQLObjectType, resolvers: dict):
    scalar_resolvers = resolvers.get(name, {})

    serialize = scalar_resolvers.get("serialize", obj.serialize)
    obj.serialize = serialize

    parse_literal = scalar_resolvers.get("parse_literal", obj.parse_literal)
    obj.parse_literal = parse_literal

    parse_value = scalar_resolvers.get("parse_value", obj.parse_value)
    obj.parse_value = parse_value
